import json
from enum import Enum
from typing import Any, List, Optional
from warnings import warn

from pydantic import Field, field_validator, model_validator
from typing_extensions import Annotated

from rhino_health.lib.dataclass import RhinoBaseModel, UIDField
from rhino_health.lib.endpoints.project.project_baseclass import WithinProjectModel
from rhino_health.lib.endpoints.user.user_baseclass import UserCreatedModel
from rhino_health.lib.endpoints.workgroup.workgroup_baseclass import PrimaryWorkgroupModel


class CodeExecutionMode(str, Enum):
    """
    A mode the CodeObject will run in
    """

    DEFAULT = "default"
    AUTO_CONTAINER_NVFLARE = "nvflare"
    AUTO_CONTAINER_SNIPPET = "snippet"
    AUTO_CONTAINER_FILE = "file"


class CodeFormat(str, Enum):
    """
    A format the code is stored and uploaded to the system
    """

    DEFAULT = "single_non_binary_file"
    S3_MULTIPART_ZIP = "s3_multipart_zip"


class RequirementMode(str, Enum):
    """
    The format the requirements are in
    """

    PYTHON_PIP = "python_pip"
    ANACONDA_ENVIRONMENT = "anaconda_environment"
    ANACONDA_SPECFILE = "anaconda_specfile"


class CodeTypes(str, Enum):
    """
    Supported CodeObject Types
    """

    GENERALIZED_COMPUTE = "Generalized Compute"
    NVIDIA_FLARE_V2_0 = "NVIDIA FLARE v2.0"
    NVIDIA_FLARE_V2_2 = "NVIDIA FLARE v2.2"
    NVIDIA_FLARE_V2_3 = "NVIDIA FLARE v2.3"
    PYTHON_CODE = "Python Code"
    INTERACTIVE_CONTAINER = "Interactive Container"


class CodeObjectBuildStatus(str, Enum):
    """
    The build status of the CodeObject
    """

    NOT_STARTED = "Not Started"
    IN_PROGRESS = "In Progress"
    COMPLETE = "Complete"
    ERROR = "Error"


class CodeObjectCreateInput(RhinoBaseModel):
    """
    @autoapi True
    Input arguments for creating CodeObject
    """

    name: str
    """@autoapi True The name of the CodeObject"""
    description: str
    """@autoapi True The description of the CodeObject"""
    input_data_schema_uids: List[str] = []
    """@autoapi True A list of uids of data schemas this CodeObject expects input datasets to adhere to."""
    output_data_schema_uids: List[Optional[str]] = []
    """@autoapi True A list of uids of data schemas this CodeObject expects output datasets to adhere to."""
    project_uid: Annotated[str, Field(alias="project")]
    """@autoapi True The CodeObject project"""
    code_type: Annotated[str, Field(alias="type")]
    """@autoapi True The code type which corresponds to the CodeTypes enum

    See Also
    --------
    rhino_health.lib.endpoints.code_object.code_object_dataclass.CodeTypes
    """
    base_version_uid: Optional[str] = ""
    """@autoapi True The first version of the CodeObject"""
    config: Optional[dict] = None
    """@autoapi True Additional configuration of the CodeObject. The contents will differ based on the model_type and code_run_type.
    
    Examples
    --------
    + CodeTypes.GENERALIZED_COMPUTE and CodeTypes.INTERACTIVE_CONTAINER
        - container_image_uri: URI of the container image to use for the model
    + CodeTypes.NVIDIA_FLARE_V2_X (existing image)
        - code_execution_mode: CodeExecutionMode
        - if CodeExecutionMode.DEFAULT requires the following additional parameters
            - container_image_uri: URI of the container image to use for the model
        - if CodeExecutionMode.AUTO_CONTAINER_NVFLARE, see CodeTypes.PYTHON_CODE below
    + CodeTypes.PYTHON_CODE
        - code_execution_mode: CodeExecutionMode = CodeExecutionMode.DEFAULT - The format the code is structured in
            - CodeTypes.PYTHON_CODE supports CodeExecutionMode.DEFAULT, CodeExecutionMode.AUTO_CONTAINER_SNIPPET, and CodeExecutionMode.AUTO_CONTAINER_FILE
            - CodeTypes.NVIDIA_FLARE_V2_X only supports CodeExecutionMode.AUTO_CONTAINER_NVFLARE
        - if CodeExecutionMode.DEFAULT requires the following additional parameters
            - python_code: str - the python code to run
        - CodeExecutionMode.AUTO_CONTAINER_SNIPPET requires the following additional parameters
            - code: str - the python code to run
        - CodeExecutionMode.AUTO_CONTAINER_FILE or CodeExecutionMode.AUTO_CONTAINER_NVFLARE
            - base_image: choose one of the following 
                - base_image_uri: str - the base docker image to use for the container
                or
                - python_version: str - the python version to use for the container
                - cuda_version: Optional[str] - the cuda version to use for the container
            - requirements_mode: Optional[RequirementMode] = RequirementMode.PYTHON_PIP - The format the requirements are in
            - requirements: List[str] - a list of requirements to install in the container, uses the pip/conda install format
            - code_format: CodeFormat the format used to pass in the code
                - CodeFormat.DEFAULT
                    - code: str - the python code to run
                - CodeFormat.S3_MULTIPART_ZIP
                    - folder_path: str | Path - the folder path to files   
                    - entry_point: str - name of the file to run first. Not used for auto container nvflare    

    See Also
    --------
    rhino_health.lib.endpoints.code_object.code_object_dataclass.CodeTypes
    rhino_health.lib.endpoints.code_object.code_object_dataclass.CodeExecutionMode
    """


class CodeObject(WithinProjectModel, CodeObjectCreateInput, UserCreatedModel):
    """
    @autoapi True
    @hide_parent_class
    A CodeObject which exists on the platform
    """

    uid: str
    """@autoapi True The unique ID of the CodeObject"""
    version: Optional[int] = None
    """@autoapi True The version of the CodeObject"""
    build_status: CodeObjectBuildStatus = CodeObjectBuildStatus.COMPLETE
    """@autoapi True The build status of the CodeObject"""
    input_data_schema_uids: Annotated[
        List[str],
        UIDField(
            model_fetcher=lambda session, uids: session.data_schema.get_data_schemas(uids),
            model_property_name="input_data_schemas",
            model_property_type="DataSchema",
            is_list=True,
        ),
    ] = []
    """@autoapi True A list of uids of data schemas this CodeObject expects input datasets to adhere to."""
    output_data_schema_uids: Annotated[
        List[Optional[str]],
        UIDField(
            model_fetcher=lambda session, uids: session.data_schema.get_data_schemas(uids),
            model_property_name="output_data_schemas",
            model_property_type="DataSchema",
            is_list=True,
        ),
    ] = []
    """@autoapi True A list of uids of data schemas this CodeObject expects output datasets to adhere to."""
    build_errors: Optional[List[str]] = None
    """@autoapi True Errors when building the CodeObject, if building in the cloud"""

    # API responses we do not want to surface to the user
    __hidden__ = ["container_image_build"]

    @property
    def build_logs(self):
        """@autoapi True Logs when building the CodeObject, if building in the cloud"""
        return self.session.code_object.get_build_logs(self.uid)

    @property
    def _code_object_built(self):
        return self.build_status in {CodeObjectBuildStatus.COMPLETE, CodeObjectBuildStatus.ERROR}

    def wait_for_build(
        self, timeout_seconds: int = 900, poll_frequency: int = 30, print_progress: bool = True
    ):
        """
        @autoapi True
        Wait for the asynchronous CodeObject to finish building

        Parameters
        ----------
        timeout_seconds: int = 900
            How many seconds to wait before timing out. Maximum of 1800.
        poll_frequency: int = 30
            How frequent to check the status, in seconds
        print_progress: bool = True
            Whether to print how long has elapsed since the start of the wait

        Returns
        -------
        code_object: CodeObject
            Dataclass representing the CodeObject
        """
        return self._wait_for_completion(
            name="code object build",
            is_complete=self._code_object_built,
            query_function=lambda code_object: code_object.session.code_object.get_code_object(
                code_object.uid
            ),
            validation_function=lambda old, new: (new.build_status and new._code_object_built),
            timeout_seconds=timeout_seconds,
            poll_frequency=poll_frequency,
            print_progress=print_progress,
        )


class CodeObjectRunInput(RhinoBaseModel):
    """
    @autoapi True
    Input parameters for running generalized code with multiple input and/or output datasets per container

    See Also
    --------
    rhino_health.lib.endpoints.code_object.code_object_endpoints.CodeObjectEndpoints.run_code_object : Example Usage
    """

    def __init__(self, *args, **kwargs):
        run_params = kwargs.get("run_params", None)
        if isinstance(run_params, dict):
            kwargs["run_params"] = json.dumps(run_params)
        secret_run_params = kwargs.get("secret_run_params", None)
        if isinstance(secret_run_params, dict):
            kwargs["secret_run_params"] = json.dumps(secret_run_params)
        super().__init__(*args, **kwargs)

    code_object_uid: str
    """@autoapi True The unique ID of the CodeObject"""
    run_params: Optional[str] = "{}"
    """@autoapi True The run params code you want to run on the datasets"""
    timeout_seconds: int = 600
    """@autoapi True The time before a timeout is declared for the run"""
    secret_run_params: Optional[str] = None
    """The secrets for the CodeObject"""
    sync: bool = False
    """@autoapi True Highly recommended to use the default parameter"""
    input_dataset_uids: List[List[str]]
    """ A list of lists of the input dataset uids.

    [[first_dataset_for_first_run, second_dataset_for_first_run ...], [first_dataset_for_second_run, second_dataset_for_second_run ...], ...] for N runs

    Examples
    --------
    Suppose we have the following CodeObject with 2 Input Data Schemas:

    - CodeObject
        + DataSchema 1
        + DataSchema 2

    We want to run the CodeObject over two sites: Applegate and Bridgestone

    The user passes in dataset UIDs for Datasets A, B, C, and D in the following order:

    [[Dataset A, Dataset B], [Dataset C, Dataset D]]

    The model will then be run over both sites with the following datasets passed to generalized compute:

    - Site A - Applegate:
        + Dataset A - DataSchema 1
        + Dataset B - DataSchema 2
    - Site B - Bridgestone:
        + Dataset C - DataSchema 1
        + Dataset D - DataSchema 2
    """
    # TODO: Remove this after figuring out how to replicate old logic
    output_dataset_names_suffix: Optional[str] = None
    """@autoapi False ..warning This argument is deprecated, will be removed soon, and does not behave properly with multiple datasets. Please use output_dataset_naming_templates"""
    output_dataset_naming_templates: Optional[List[str]] = None
    """ A list of string naming templates used to name the output datasets at each site.
    You can use parameters in each template surrounded by double brackets ``{{parameter}}`` which will
    then be replaced by their corresponding values.

    Parameters
    ----------
    workgroup_name:
        The name of the workgroup the CodeObject belongs to.
    workgroup_uid:
        The name of the workgroup the CodeObject belongs to.
    model_name:
        The CodeObject name
    input_dataset_names.n:
        The name of the nth input dataset, (zero indexed)

    Examples
    --------
    Suppose we have two input datasets, named "First Dataset" and "Second Dataset"
    and our CodeObject has two outputs:

    | output_dataset_naming_templates = [
    |     "{{input_dataset_names.0}} - Train",
    |     "{{input_dataset_names.1}} - Test"
    | ]

    After running Generalized Compute, we will save the two outputs as "First Dataset - Train" and "Second Dataset - Test"
    """

    @field_validator("output_dataset_names_suffix")
    @classmethod
    def warn_on_usage_of_obsolete_arg(cls, value):
        if value is not None:
            warn(
                "output_dataset_names_suffix is deprecated and will be removed very soon, please use output_dataset_naming_templates instead",
                RuntimeWarning,
            )
        return value

    @model_validator(mode="after")
    def max_timeout_check(self):
        """
        @autoapi False
        """
        if self.sync == True and self.timeout_seconds > 600:
            raise ValueError(
                "Timeout seconds cannot be greater than 600 when run in synchronous mode"
            )
        return self


class CodeRunWaitMixin(RhinoBaseModel):
    """
    @autoapi False
    """

    code_run_uid: Optional[str] = None
    """
    @autoapi True The UID of the model result
    """
    _code_run: Any = None

    def wait_for_completion(self, *args, **kwargs):
        """
        @autoapi True
        Wait for the asynchronous Code Run to complete, convenience function call to the same function
        on the CodeRun object.

        Returns
        -------
        code_run: CodeRun
            Dataclass representing the CodeRun of the CodeObject

        See Also
        --------
        rhino_health.lib.endpoints.code_run.code_run_dataclass.CodeRun: Response object
        rhino_health.lib.endpoints.code_run.code_run_dataclass.CodeRun.wait_for_completion: Accepted parameters
        """
        if not self._code_run:
            self._code_run = self.session.code_run.get_code_run(self.code_run_uid)
        if self._code_run._process_finished:
            return self._code_run
        self._code_run = self._code_run.wait_for_completion(*args, **kwargs)
        return self._code_run


class CodeObjectRunResponse(CodeRunWaitMixin):
    """
    @autoapi False
    """

    status: str
    """
    @autoapi True The status of the run
    """
    output_dataset_uids: Optional[List[str]] = []
    """
    @autoapi True A list of output dataset uids for the run
    """

    @property
    def code_run(self):
        """
        @autoapi True
        Return the CodeRun associated with this CodeObject

        .. warning:: The result of this function is cached.
            Be careful calling this function after making changes

        Returns
        -------
        code_run: CodeRun
            Dataclass representing the CodeRun
        """
        if self._code_run:
            return self._code_run
        if self.code_run_uid:
            self._code_run = self.session.code_run.get_code_run(self.code_run_uid)
            return self._code_run
        else:
            return None


class CodeObjectRunAsyncResponse(CodeObjectRunResponse):
    task_uids: List[str]

    @property
    def code_run(self):
        return self.wait_for_completion()


class CodeObjectRunSyncResponse(CodeObjectRunResponse):
    errors: Optional[List[Any]] = []
    warnings: Optional[List[Any]] = []

    def wait_for_completion(self, *args, **kwargs):
        return self.code_run


## NVFlare


class ModelTrainInput(RhinoBaseModel):
    """
    @autoapi True
    Input for training an NVFlare Model

    See Also
    --------
    rhino_health.lib.endpoints.code_object.code_object_endpoints.CodeObjectEndpoints.train_model : Example Usage
    """

    code_object_uid: str
    """The unique ID of the CodeObject"""
    input_dataset_uids: List[str]
    """A list of the input Dataset uids"""
    validation_dataset_uids: List[str]
    """A list of the Dohort uids for validation"""
    validation_datasets_inference_suffix: str
    """The suffix given to all output datasets"""
    simulate_federated_learning: Annotated[bool, Field(alias="one_fl_client_per_dataset")]
    """Run simulated federated learning on the same on-prem installation by treating each dataset as a site"""
    config_fed_server: Optional[str] = None
    """The config for the federated server"""
    config_fed_client: Optional[str] = None
    """The config for the federated client"""
    secrets_fed_server: Optional[str] = None
    """The secrets for the federated server"""
    secrets_fed_client: Optional[str] = None
    """The secrets for the federated client"""
    timeout_seconds: int
    """The time before a timeout is declared for the run"""


class ModelInferenceAsyncResponse(CodeRunWaitMixin):
    @property
    def code_run(self):
        return self.wait_for_completion()


class ModelTrainAsyncResponse(CodeRunWaitMixin):
    """
    Response of training an NVFlare Model
    .. warning:: This feature is under development and the interface may change
    """

    status: str
    """
    @autoapi True The status of the run
    """

    @property
    def code_run(self):
        """
        @autoapi True
        Return the CodeRun associated with the NVFlare training

        .. warning:: The result of this function is cached.
            Be careful calling this function after making changes

        Returns
        -------
        code_run: CodeRun
            Dataclass representing the CodeRun
        """
        return self.wait_for_completion()
